"""Helper methods for testing."""
from typing import List
from fastapi import HTTPException

from ..structs.landmark import Landmark


def landmarks_to_osmid(landmarks: List[Landmark]) -> List[int] :
    """
    Convert the list of landmarks into a list containing their osm ids for quick landmark checking.
    
    Args :
        landmarks (list): the list of landmarks
    
    Returns :
        ids (list)      : the list of corresponding OSM ids
    """
    ids = []
    for landmark in landmarks :
        ids.append(landmark.osm_id)

    return ids

def fetch_landmark(client, landmark_uuid: str):
    """
    Fetch landmark data from the API based on the landmark UUID.

    Args:
        landmark_uuid (str): The UUID of the landmark.

    Returns:
        dict: Landmark data fetched from the API.
    """
    response = client.get(f"/landmark/{landmark_uuid}")

    if response.status_code != 200:
        raise HTTPException(status_code=999,
                            detail=f"Failed to fetch landmark with UUID {landmark_uuid}: {response.status_code}")

    json_data = response.json()

    if "detail" in json_data:
        raise HTTPException(status_code=999, detail=json_data["detail"])


    return json_data


def load_trip_landmarks(client, first_uuid: str) -> List[Landmark]:
    """
    Load all landmarks for a trip using the response from the API.

    Args:
        first_uuid (str) : The first UUID of the landmark.

    Returns:
        landmarks (list) : An list containing all landmarks for the trip.
    """
    landmarks = []
    next_uuid = first_uuid

    while next_uuid is not None:
        landmark_data = fetch_landmark(client, next_uuid)
        # # Convert UUIDs to strings explicitly
        # landmark_data = {
        #     key: str(value) if isinstance(value, UUID) else value
        #     for key, value in landmark_data.items()
        # }
        landmarks.append(Landmark(**landmark_data)) # Create Landmark objects
        next_uuid = landmark_data.get('next_uuid')  # Prepare for the next iteration

    return landmarks


def log_trip_details(request, landmarks: List[Landmark], duration: int, target_duration: int) :
    """
    Allows to show the detailed trip in the html test report.
    
    Args:
        request:
        landmarks (list): the ordered list of visited landmarks 
        duration (int): the total duration of this trip
        target_duration(int): the target duration of this trip
    """
    trip_string = [f"{landmark.name} ({landmark.attractiveness} | {landmark.duration}) - {landmark.time_to_reach_next}" for landmark in landmarks]

    # Pass additional info to pytest for reporting
    request.node.trip_details = trip_string
    request.node.trip_duration = str(duration)              # result['total_time']
    request.node.target_duration = str(target_duration)
